#! /usr/bin/python2

# $Id: algorithm.py,v 1.4 2003/06/25 09:59:40 ensc Exp $
# Copyright (C) 2001 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
#  
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#  

class DFS:
    """An class doing a depth-first-search (DFS). This is an abstract
    class an the following methods must be implemented by specialized
    child-classes:

    calculateAdjacencies(node) ...  returns a list of a nodes which can
                                    be reached from node


    There exists a warnRecursion() method which gets called when a
    cycle was detected."""

    class marker:
        """Stores information about a depth-first-search (DFS)."""
    
        def __init__(this):
            this.__time  = 0
            this.__visit = {}
            this.__leave = {}
            this.__mark  = {}
    
        def visit(this, node):
            """Called when the node is visited the first time. It will
            throw an exception if the node was already visited."""
            
            assert(not this.isSeen(node))
    
            this.__time = this.__time + 1
            val = this.calculateNodeValue(node)
    
            this.__mark[val]  = 1
            this.__visit[val] = this.__time
    
        def leave(this, node):
            """Called after all the nodes below node were visited. It
            will throw an exception if the node was leaved already or not
            visited yet."""
            
            assert(this.isVisited(node))
    
            this.__time = this.__time + 1
            val = this.calculateNodeValue(node)
            
            del this.__mark[val]
            this.__leave[val] = this.__time
    
        def isVisited(this,node):
            """Tells if we are visiting nodes below node currently. This
            means visit() but not leave() was called already.
    
            Returning true indicates a cycle in the DFS."""
            
            val = this.calculateNodeValue(node)
            return this.__mark.get(val,0)
    
        def isLeaved(this, node):
            """Tells if the node and all nodes below node were visited
            already. This means leave() was called."""
    
            val = this.calculateNodeValue(node)
            return this.__leave.get(val,0)>0
    
        def isSeen(this, node):
            """Tells if the node was seen already. This means visit() was
            called already."""
            
            val = this.calculateNodeValue(node)
            return this.__visit.get(val,0)>0
    
        def getInfo(this, node):
            """Returns the times when the node was visited and leaved."""
            
            val = this.calculateNodeValue(node)
            return (this.__visit.get(val), this.__leave.get(val))

        def getSortedLeaveList(this):
            """Returns the visited elements as a list which is sorted
            by their leave-time. This can be used e.g. by a topological
            sort."""
            
            return this.__getSortedMapKeys(this.__leave)

        def getSortedVisitList(this):
            """Returns the visited elements as a list which is sorted
            by their visit-time."""
            return this.__getSortedMapKeys(this.__visit)

        def __getSortedMapKeys(this, m):
            lst = m.items()
            lst.sort(lambda x,y: x[1]-y[1])

            return map(lambda x: x[0], lst)

        
    def __init__(this, obj):
        """Initializes the class so a call of visit() would begin the
        DFS beginning at object obj."""
            
        this.reinit(obj)

    def reinit(this,obj):
        """When calling this method, the next DFS will be started at
        node <obj>."""
        
        this.__obj = obj

    def resort(this, objs):
        """Resorts the list objs so that subsequent DFSs with the
        elements of this list can reuse information from previous
        runs."""

        return objs
    
    def visitWrapper(this, pre_action, post_action, extras, dfs=None):
        """Executes the DFS beginning at the 'obj' object given to
        the constructor. Before a cycle goes into the depth the
        pre_action will be called. The post_action will be called
        after the recursive call returns.
    
        Both actions are methods called with three argument:
    
        obj        - the object being visited currently
        extras     - the extras argument
        dfs_marker - an DFSMarker object representing the current
                     state of the DFS"""
            
        if dfs==None: dfs = this.getDFSMarker()
        this.__visit(pre_action, post_action, dfs, this.__obj, extras)

        return dfs

    def getStartObject(this):
        """Returns the object from which the DFS was (or will be)
        started."""
        return this.__obj
    
    def calculateAdjacencies(this,obj):
        """Calculates and returns the nodes which can be reached from
        obj. The result must be a list or tuple whose elements are
        from the same type as obj."""
        
        raise AttributeError("calculateAdjacencies")
    
    def warnRecursion(this, obj, dfs_marker):
        """Gets called when a recursion was detected in the current
        DFS. It gets called with the node obj where the recursion
        happened and the current state dfs_marker."""

        pass
    
    def __visit(this, pre_action, post_action, dfs_marker, obj, extra):
        if dfs_marker.isSeen(obj):
            if dfs_marker.isVisited(obj): this.warnRecursion(obj, dfs_marker)
            return
        
        dfs_marker.visit(obj)

        if pre_action: pre_action(obj, extra, dfs_marker)
        
        for i in this.calculateAdjacencies(obj):
            this.__visit(pre_action, post_action, dfs_marker, i, extra)
            
        if post_action: post_action(obj, extra, dfs_marker)
            
        dfs_marker.leave(obj)


### Local Variables:
### mode: python
### End:    
