#! /usr/bin/python2

# $Id: rpmDirectoryCheck.py,v 1.16 2003/07/29 13:07:42 ensc Exp $
# Copyright (C) 2001 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
#  
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#  

import os
import string
import getopt, sys
import config

from config import cfg
import rpmWrapper
import algorithm
import debug
import params

from tools import unique

class abstractRegistry:
    """Class which can be used to register <key,value> pairs where
    more than one value can be assigned to a key.

    This an abstract class and should be derived."""
    
    def __init__(this):
        """Constructs the object."""
        
        this.__values = {}

    def add(this, key, value):
        """Registers the <key,value> pair."""

        if not this.__values.has_key(key): this.__values[key] = []
        this.__values[key].append(value)

    def isRegistered(this, key):
        """Says if a <k,value> pair exists where k==key."""
        
        return this.__values.has_key(key)

    def getAsMap(this):
        """Returns the stored data as a {key -> values} map where
        values is a list of objects associated with key."""

        return this.__values;
        
    def __len__(this):
        return len(this.__values)
    
    def __repr__(this):
        return `this.__values`


class directoryRegistry(abstractRegistry):
    """Specialized child-class of abstractRegistry."""
    
    def __init__(this):
        abstractRegistry.__init__(this)
        this.add('/', None)

    def add(this, dir, package):
	assert(type(dir) == type(""))
	assert(package==None or isinstance(package, rpmWrapper.package))

	return abstractRegistry.add(this, dir, package)


class rpmDFSVisitor(algorithm.DFS):
    """Specialized child-class of algorithm.DFS which uses the
    rpm.package class to gain information about adjcencies of an
    rpm-package."""
    
    class rpmDFSMarker(algorithm.DFS.marker):
        """Specialized child-class of algorithm.DFS.marker which uses
        knowledge about the rpmWrapper.package class."""
        
        def __init__(this):
            algorithm.DFS.marker.__init__(this)
    
        def calculateNodeValue(this, package):
            assert(isinstance(package,rpmWrapper.package))
            
            return package.getId()

    def __init__(this, rpm_id, rpm_registry):
        this.__rpms = rpm_registry
        algorithm.DFS.__init__(this, rpm_id)
        
    def reinit(this, rpm_id):
        if rpm_id==None:                            tmp = None
        elif isinstance(rpm_id,rpmWrapper.package): tmp = rpm_id
        else:                                       tmp = this.lookupPackage(rpm_id)
            
        algorithm.DFS.reinit(this, tmp)

    def lookupPackage(this, rpm_id):
        tmp     = this.__rpms.get(rpm_id)
        tmp_pkg = None

        if tmp==None:
            tmp_pkg  = rpmWrapper.package(rpm_id)
            rpm_id   = tmp_pkg.getId()
            tmp      = this.__rpms.get(rpm_id)

        if tmp==None:
            if tmp_pkg==None: tmp = rpmWrapper.package(rpm_id)
            else:             tmp = tmp_pkg

            this.__rpms[rpm_id] = tmp

        return tmp
                  
    def warnRecursion(this, obj, dfs_marker):
        #print >> sys.stderr, "Warning: loop at %s" % str(obj)
        pass

    def getRPMRegistry(this): return this.__rpms
    
    def getDFSMarker(this):
        return rpmDFSVisitor.rpmDFSMarker()

    def calculateAdjacencies(this, obj):
        assert(isinstance(obj, rpmWrapper.package))

        result = []
        for i in obj.getDependencies():
            tmp = this.lookupPackage(i)
            result.append(tmp)

        return result


class baseReport:
    """Abstract class defining interfaces for a report class."""
    
    def __init__(this): None

    def execute(this): raise AttributeError("execute")


class directoryChecker(rpmDFSVisitor):
    """This class makes a DFS and checks if directories are owned by
    an rpm-package or its requirements. Its visit() method takes a
    list of packages to visit. To gain maximum performance, this list
    should be sorted so that packages having fewest dependencies are
    visited first.

    The resort() method executes a topological-sort to gain such a
    list."""



    class orphanedReport(baseReport):
        """Specialized report using knowledge about the rpmWrapper.package
        class. It outputs gained information as XML."""
        
        def __init__(this, owner, dir_registry):
            from xml.dom.minidom import Document

            this.__owner    = owner
            this.__bad_rpms = {}
            this.__all_dirs = dir_registry
            this.__doc      = Document()
        
        def registerOrphaned(this, package, path):
            """It registers the <path> as not covered by <package> (or
            its requirements)."""
            
            assert(isinstance(package, rpmWrapper.package))
            assert(type(path)==type(""))

            id = package.getId()
            if not this.__bad_rpms.has_key(id):
                this.__bad_rpms[id] = { 'paths' : [],
                                        'rpm'   : package }

            this.__bad_rpms[id]['paths'].append(path)

        def __addChild(this, node, name):
            """XML-helper method creating a child with name <name> to
            the <node>."""
            
            n = this.__doc.createElement(name)
            node.appendChild(n)

            return n

        def __addChildText(this, node, name, text):
            """XML-helper method creating a child-node of <node>. This
            child-node is a text-node with name <name> and data <text>."""
            
            n = this.__doc.createElement(name)
            node.appendChild(n)
            n.appendChild(this.__doc.createTextNode(text))

            return n

        def isSpecialPath(this, path):
            """Says if the path <path> is a special one. Currently
            only '/usr/share/locale' is handled as a such one."""

            for i in params.SPECIAL_PATHS_FULL:
                if path==i: return 1

            for i in params.SPECIAL_PATHS:
                if string.find(path, i) == 0: return 1

            return 0
	
        def execute(this):
            """Generates the XML output."""

            all_paths = this.__all_dirs.getAsMap()
            all_rpms  = this.__owner.getRPMRegistry().items()
            
            n_doc     = this.__addChild(this.__doc, 'packagereport')
            n_docinfo = this.__addChild(n_doc, 'documentinfo')
            this.__addChildText(n_docinfo, 'generator',        config.PACKAGE)
            this.__addChildText(n_docinfo, 'generatorversion', config.VERSION)

            progress = debug.progressMeter("Evaluating information about %i packages..." % len(all_rpms))
            
            for (id,package) in all_rpms:
                progress.inc()
                
                n_pnode = this.__addChild(n_doc, 'package')
                n_pnode.setAttribute('id', package.getData().fullname)
                this.__addChildText(n_pnode, 'name',     package.getData().name)
                this.__addChildText(n_pnode, 'version',  package.getData().version)
                this.__addChildText(n_pnode, 'release',  package.getData().release)

                n_reqs  = this.__addChild(n_pnode, 'requirements')
                n_badps = this.__addChild(n_pnode, 'badpaths')

                for i in package.getDependencies():
                    pkg = this.__owner.lookupPackage(i)
                    this.__addChildText(n_reqs, 'id', pkg.getData().fullname)

                bad_rpm = this.__bad_rpms.get(id, None)

                if bad_rpm!=None:
                    paths   = unique(bad_rpm['paths'])

                    for i in paths:
                        tmp = all_paths.get(i)
                        assert(tmp==None or type(tmp)==type([]))
                        if tmp==[None,]: continue
                        
                        n_badp = this.__addChild(n_badps, 'badpath')
                        this.__addChildText(n_badp, 'id', i)

                        if this.isSpecialPath(i):
                            n_badp.setAttribute('type', 'special')
                        elif not tmp:
                            n_badp.setAttribute('type', 'unowned')
                            
                        if tmp:
                            n_owners = this.__addChild(n_badp, 'owners')
                            for owner in tmp:
                                if owner==None: continue
                                this.__addChildText(n_owners, 'id', owner.getData().fullname)

            del progress
            
            this.__doc.normalize()
            print this.__doc.toxml()
    
    def __init__(this, start_rpm, rpm_registry):
        """Constructs a directory-checker object starting at
        <start_rpm> and using <rpm_registry> as a place where
        generated rpmWrapper.package are stored. Latter aspect is
        for performance-purposes only."""
        
        assert(start_rpm==None or isinstance(start_rpm, rpmWrapper.package))

        rpmDFSVisitor.__init__(this, start_rpm, rpm_registry)

        this.__all_paths         = directoryRegistry()
        this.__registered_paths  = {}
        this.__report            = directoryChecker.orphanedReport(this, this.__all_paths)

    def doPostCheck(this, package, registry):
        """Called after the DFS started at <package> is
        finished. <registry> is a map where each key specifies a path
        used by the <package> or its childs. When a file contained in
        <package> uses an unknown path, it will be registered by the
        report-object."""
        
        assert(isinstance(package, rpmWrapper.package))

        paths = package.getFiles().getPaths()
        for i in paths:
            if registry.has_key(i): continue
            this.__report.registerOrphaned(package, i)

        
    def registerDirectories(this, package, registry, dummy):
        """Registers all directories owned by package <package> or its
        child into the <registry>-map and the object-global __all_paths
        registry.

        If the object-global __registered_paths map has the key <package>,
        the directories stored there are used. Else they are taken from
        <package> by calling its getFiles().getDirectories() method.

        This strategy uses cached values of already visited nodes and
        increases performance when calculateAdjacencies() skips already
        visited child-nodes.


        This method is called as a <post_action> function of the
        DFS-visitWrapper."""

        assert(isinstance(package, rpmWrapper.package))
        
        id   = package.getId()
        dirs = this.__registered_paths.get(id)

        if dirs!=None:
            for path in dirs.keys(): registry[path] = 1
        else:
            dirs = package.getFiles().getDirectories()
            func = this.__all_paths.add
            for i in dirs:
                dirname = i.getName() 
                registry[dirname] = 1
                func(dirname, package)

    def calculateAdjacencies(this, package):
        """Returns adjacencies of <package> or an empty list if a DFS
        at <package> was executed already."""
        
        tmp = this.__registered_paths.has_key(package.getId())
        if tmp: return ()
        else:   return rpmDFSVisitor.calculateAdjacencies(this, package)
        
    def visit(this, package):
        """Creates a directory registry and calls visitWrapper() with
        registerDirectories as the post_action method."""

        if type(package) in (type(()), type([])):
            count = 0

            for i in package:
                count   = count+1
                percent = count*100/len(package)
                if   cfg.verbosity>=2: print >> sys.stderr, '%-50s\t[%4u/%u  %6s]' % (i,count, len(package),'(%1.0f%%)' % percent)
                elif cfg.verbosity==1: sys.stderr.write('.')
                
                this.visit(i)
            
            return
            
        this.reinit(package)
        
        dirs   = {}
        result = rpmDFSVisitor.visitWrapper(this, None, this.registerDirectories, dirs)
        this.__registered_paths[ this.getStartObject().getId() ] = dirs

        this.doPostCheck(this.getStartObject(), dirs)
        
        return result

    def getReport(this):
        """Returns the orphanedReport object telling gained results."""
        
        return this.__report

    def resort(this, rpms, complete_all=0):
        """Executes a topological sort on all packages in a <rpms>
        list. When complete_all is true all direct or indirect
        requirements found while this sort are returned also. Else
        only the sorted <rpms> are returned.

        The elements of the returned list are always full names of the
        <rpms>. """
        
        deps  = {}
        count = 0

        progress = debug.progressMeter('Resorting %i packages...' % (len(rpms)))
        dfs = this.getDFSMarker()

        id    = []
        for i in rpms:
            progress.inc()
            this.reinit(i)
            id.append(this.getStartObject().getId())
            dfs = rpmDFSVisitor.visitWrapper(this, None, None, None, dfs)

        result = []
        for i in dfs.getSortedLeaveList():
            if complete_all or (i in id):
                result.append(this.lookupPackage(i))
            
        del progress
        
        return result

def usage(cmd):
    print """%s %s
Copyright (C) 2001 - Enrico Scholz
This program may be freely redistributed under the terms of the GNU GPL

%s [-h|--help] [-q|--quiet] [-v|--verbose] [--resort-only] [--dont-resort] [--dbpath <path>] [-c|--complete] -a|--all|-f <filename>|<packages>+

-h|--help       ...  help
-v|--verbose    ...  increase verbosity; multiple -v options are supported
                     [default: verbosity level 2]
-q|--quiet      ...  decrease verbosity; multiple -q options are supported;
                     this is the exact opposite of hte -v option
-f              ...  fetch package-list from file <filename>
-c|--complete   ...  determine badpaths in each package which is a direct or
                     indirect requirement of the given package(s). This
                     option conflicts with --dont-resort

--resort-only   ...  resort given packages and gives them out on stdout, but
                     do not process them. If used in combination with the
                     --complete option the new packages determined there are
                     output also
--dont-resort   ...  skip the resorting-step; this can slow down processing

--dbpath <path> ...  use rpm-db located in <path>

-a|-all         ...  visit all packages in the rpm-db
<packages>+     ...  packages to visit
    """ % (config.PACKAGE, config.VERSION, cmd)

def main(cfgdir):
    import gc
    gc.disable()
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvf:qca",
                                   ["help", "verbose", "resort-only", "dont-resort",
                                    "quiet", "complete", "dbpath=", "all", "version"])
    except getopt.GetoptError:
        usage(sys.argv[0])
        sys.exit(2)

    output       = None
    resort_only  = 0
    dont_resort  = 0
    visit_all    = 0
    complete_all = 0
    
    for o, a in opts:
        if o in ("-h", "--help"):
            usage(sys.argv[0])
            sys.exit()
        elif o in ("-v", "--verbose"):
            cfg.verbosity.inc()
        elif o in ("-q", "--quiet"):
            cfg.verbosity.dec()
        elif o in ("-f",):
            fd = open(a)
            ## TODO: error-handling
            args.extend(map(lambda x: string.replace(x,"\n",''), fd.readlines()))
            fd.close()
        elif o in ("--resort-only",):
            resort_only = 1
        elif o in ("--dont-resort",):
            dont_resort = 1
        elif o in ("--dbpath",):
            config.cfg.dbpath = a
        elif o in ("-c", "--complete"):
            complete_all = 1
        elif o in ("-a", "--all"):
            visit_all = 1
        elif o in ("--version",):
            print "%s %s" % (config.PACKAGE, config.VERSION)
            sys.exit()

    params.init(cfgdir)

    if visit_all:
        args.extend( rpmWrapper.getAllPackages() )

    if len(args)==0:
        usage(sys.argv[0])
        sys.exit("You must give at least one package\n")

    if output==None: output = 'report'

    dfs_class = directoryChecker

    rpm_cache = {}
    dfs       = dfs_class(None, rpm_cache)

    if not dont_resort: args = dfs.resort(args, complete_all)

    if resort_only:
        for i in args: print i
        sys.exit()

    dfs.visit(args)
    dfs.getReport().execute()
    
    debug.printMemInfo()

    if cfg.verbosity>0 and rpmWrapper.ambiguous_deps:
        #print map(lambda x: tuple(x), rpmWrapper.ambiguous_deps)
        print >>sys.stderr, "\nWarning: the following packages are providing the same libraries which may cause wrong result:\n"
        print >>sys.stderr, reduce(lambda x,y: "%s  %s\n" % (x,y), unique(map(lambda x: tuple(x), rpmWrapper.ambiguous_deps)), "")
        print >>sys.stderr, "To fix this, define preferences in %s/dircheck.conf by adding entries to the RPM_PREFS array" % cfgdir

if __name__=="__main__":
    main(".")


### Local Variables:
### mode: python
### End:    
